###########################################
#
# Example script to build a pinout diagram
#
###########################################

from pinout import diagram

# Create a new diagram
pinout_diagram = diagram.Diagram()

# Add a stylesheet
pinout_diagram.add_stylesheet('tutorial_pinout_styles.css', embed=True)

# Add a hardware image
pinout_diagram.add_image(0, 0, 220, 300, 'tutorial_pinout_hardware.png', embed=True)

# Set some default values for labels
diagram.Label.default_width = 70
diagram.Label.default_height = 26
diagram.Label.default_gap = 6
diagram.Label.default_cnr = 3

# Document pin and label data
pin_headers = [
    {
        # Bottom left
        'pin_coords': (65, 284),
        'label_coords': (-30 ,320),
        'pitch': 30,
        'labels': [
            [('3', 'gpio'),('A3', 'analog')], 
            [('4', 'gpio'), ('PWM','pwm', 70, 26, 76)],
        ]
    },{
        # Bottom right
        'pin_coords': (125, 284),
        'label_coords': (250, 320),
        'pitch': 30,
        'labels': [
            [('5', 'gpio'), ('A1', 'analog')], 
            [('3V3', 'pwr-mgt')],
        ]
    },{
        # Left
        'pin_coords': (15, 105),
        'label_coords': (-30, 105),
        'pitch': 30,
        'labels':[
            [('Vcc', 'pwr-mgt')], 
            [('RST', 'pwr-mgt')],  
            [('1', 'gpio'), ('CLK','comms'), ('PWM','pwm')], 
            [('2', 'gpio'),('A8', 'analog'), ('PWM','pwm')]
        ]
    },{
        # Right
        'pin_coords': (205, 105),
        'label_coords': (250, 105),
        'pitch': 30,
        'labels':[
            [('GND', 'pwr-mgt')], 
            [('8', 'gpio qqqq'),('DAC2', 'analog-out'), ('PWM','pwm')], 
            [('7', 'gpio'),('A1', 'analog'), ('MISO','comms')], 
            [('6', 'gpio'),('A2', 'analog'), ('MOSI','comms')],
        ]
    },{
        # Aux1
        'pin_coords': (46, 78),
        'label_coords': (-30, -20),
        'labels': [
            [('9', 'gpio'),('LED', 'led')],
        ]
    },{
        # Aux2
        'pin_coords': (62, 95),
        'label_coords': (-30, -50),
        'labels': [
            [('AREF', 'pwr-mgt')],
        ]
    },{
        # LED
        'pin_coords': (87, 85),
        'label_coords': (110, -50),
        'labels': [
            [('*LED', 'led')],
        ]
    }
]

# Alternative: Single Pins can be added via diagram.add_pin 
# pinout_diagram.add_pin(87, 85, 110, -50, [('LED', 'led')])

# Add data into pinout_diagram
for header in pin_headers:
    pinout_diagram.add_pin_header(header)

# Export the finished diagram
pinout_diagram.export('tutorial_pinout_diagram.svg', overwrite=True)


from jinja2 import Environment, PackageLoader, select_autoescape


env = Environment(
    loader=PackageLoader('pinout','templates'),
    autoescape=select_autoescape(['html', 'xml']),
)

env.trim_blocks = True
env.lstrip_blocks = True

# Base SVG elements
svg = env.get_template('svg.svg')
svg_group = env.get_template('group.svg')
svg_image = env.get_template('image.svg')
svg_style = env.get_template('style.svg')

# CSS styles
stylesheet = env.get_template('stylesheet.css')

# Component SVG templates (made from multiple svg elements)
svg_legend = env.get_template('legend.svg')
svg_label = env.get_template('label.svg')
svg_pin = env.get_template('pin.svg')



import base64
from itertools import zip_longest
from pathlib import Path
from collections import namedtuple
from .templates import svg_group, svg_image, svg_legend, svg_style, svg_label, svg_pin


_BoundingBox = namedtuple('_BoundingBox',('x y w h'))
_Rect = namedtuple('_Rect',['x','y','w','h','r'])
_Line = namedtuple('_Line',('x1','x2','y1','y2'))
_Coords = namedtuple('_Coords',('x y'))



class Label:
    """Create an individual label. Labels must be associated to a Pin to be located and rendered in the final diagram. Pin.add_label() is the recommended method of creating labels.

    :param name: Text that appear on the label.
    :type name: str
    :param tags: Applied to the label as css class selector(s). Multiple tags can be included as a space separated list.
    :type name: str
    :param width: Width of the label rectangle.
    :type width: int
    :param height: Height of the label rectangle.
    :type height: int
    :param gap: Space between the label rectangle and proceeding label or pin location. The gap contains a graphical 'leader-line'.
    :type gap: int
    :param cnr: Corner radius or teh label rectangle.
    :type cnr: int
    """
    
    #:Default label box width.
    default_width = 70

    #:Default label box height.
    default_height = 25

    #:Default label gap.
    default_gap = 5

    #:Default label box corner-radius.
    default_cnr = 2

    def __init__(self, name, tags, width=None, height=None, gap=None, cnr=None):
        """ Constructor method
        """
        self.name = name
        self.tags = tags.strip()
        self._width = width
        self._height = height
        self._gap = gap
        self._cnr = cnr

    @property
    def width(self):
        return self._width or Label.default_width

    @property
    def height(self):
        return self._height or Label.default_height

    @property
    def gap(self):
        return self._gap or Label.default_gap

    @property
    def cnr(self):
        return self._cnr or Label.default_cnr


class Pin:
    def __init__(self, pin_x, pin_y, label_x=None, label_y=None, label_tuples=None):
        """Each Pin documents a location in the diagram and manages position and rendering of labels associated to it.

        :param pin_x: Location of the pin on the x axis 
        :type pin_x: int
        :param pin_y: location of the pin on the y axis
        :type pin_y: int
        :param label_x: X coordinate of the labels relative to the diagram's (0,0) coordinate.
        :type label_x: int
        :param label_y: Y coordinate of the labels relative to the diagram's (0,0) coordinate
        :type label_y: int
        :param label_tuples: A list of tuples can be supplied to streamline the pin and label creation process into a single step. Each tuple must represent the required arguments of Pin.add_label(). Defaults to None
        :type label_tuples: List, optional
        """
        self.pin_coords = _Coords(pin_x, pin_y)
        # Label_coords relative to pin_coords
        self.label_coords = _Coords(label_x - pin_x, label_y - pin_y) 
        self.labels = []
        if label_tuples:
            for label in label_tuples:
                self.add_label(*label)
    
    def add_label(self, name, tags=None, width=None, height=None, gap=None):
        """Add a label to the pin.

        :param name: Text that appear on the label
        :type name: str
        :param tags: Applied to the label as css class selector(s). Multiple tags can be included as a space separated list, defaults to None
        :type tags: str, optional
        :param width: Width of the label rectangle.
        :type width: int
        :param height: Height of the label rectangle.
        :type height: int
        :param gap: Space between the label rectangle and proceeding label or pin location. The gap contains a graphical 'leader-line'.
        :type gap: int
        """
        self.labels.append(Label(name, tags, width, height, gap))
    
    @property
    def width(self):
        """The total width of each pin is determined by the collective widths of all labels associated with the pin.

        :return: Sum of all label widths
        :rtype: int
        """       
        return sum([label.width + label.gap for label in self.labels]) + abs(self.label_coords.x)

    @property
    def height(self):
        """Each label associated with a pin can have its height independently set. The overall height of the pin is thus dictated by its tallest labels.

        :return: Height of the tallest label.
        :rtype: int
        """
        return abs(self.label_coords.y) + max([label.height or Label.default_height for label in self.labels]) - self.tallest_label / 2

    @property
    def tallest_label(self):
        """Finds and returns the height of the tallest label associated with a pin.

        :return: Height of the tallest label.
        :rtype: int
        """
        return max([l.height for l in self.labels])

    @property
    def bounding_box(self):
        """Calculate a rectangular box that documents the bounds and location the rendered object 

        :return: namedTuple documenting x, y, width, and height.
        :rtype: diagram._BoundingBox
        """
        if self.label_coords.x > 0:
            # labels located right of pin
            x = self.pin_coords.x
        else:
            # labels located left of pin
            x = self.pin_coords.x - self.width

        if self.label_coords.y > 0:
            # labels located below pin
            y = self.pin_coords.y
        else:
            # labels located above pin
            y = self.pin_coords.y + self.label_coords.y - self.tallest_label / 2

        return _BoundingBox(x, y, self.width, self.height)
        

    def render(self):
        """Generates SVG tags of all associated labels.

        :return: SVG components
        :rtype: str
        """
        output = ''
        for i, label in enumerate(self.labels):
            tags = ('label ' + label.tags).strip()
            
            offset = sum(l.width + l.gap for l in self.labels[:i])
            output += svg_label.render(
                selectors = ' '.join(['label', label.tags]),
                leaderline_class = label.tags.split(' ')[0],
                label = label,
                offset = offset,
                flip = self.label_coords.x < 0,
            )
            
        return  svg_pin.render(
            x = self.pin_coords.x,
            y = self.pin_coords.y,
            label_coords = self.label_coords,
            leaderline = f'M 0 0 V {self.label_coords.y} H {self.label_coords.x}',
            leaderline_class = self.labels[0].tags.split(' ')[0],
            flip = self.label_coords.x < 0,
            content = output,
            selectors = 'pin'
        )


class Image:
    def __init__(self, x, y, width, height, filepath, embed=False):
        """Include an image in the diagram.

        :param x: Location of the image on the x axis
        :type x: int
        :param y: Location of the image on the y axis
        :type y: int
        :param width: Width of image in the diagram (may differ from actual image width)
        :type width: int
        :param height: Height of the image in the diagram (may differ from actual image height)
        :type height: int
        :param filepath: Filename, including path, to the image. Relative paths are relative to the current working directory.
        :type filepath: string
        :param embed: Elect to link or embed an external image. Embedded images are base64 encoded. Default to False.
        :type embed: bool
        """
        self.x = x
        self.y = y
        self.path = filepath
        self.width = width
        self.height = height
        self.embed = embed

    @property
    def bounding_box(self):
        """Calculate a rectangular box that documents the bounds and location the rendered object 

        :return: namedTuple documenting x, y, width, and height.
        :rtype: diagram._BoundingBox
        """
        return _BoundingBox(self.x, self.y, self.width, self.height)
    
    def render(self):
        """Generates SVG <image> tag using the image 'filename', Note that JPG and PNG are the only binary images files officially supported by the SVG format. If 'embed' is True the image is assigned to the path as a data URI. JPG and PNG image are base64 encoded, SVG files included verbatim. Otherwise the path 'src' is assigned 'filename'. Note: 'filename' includes the path to the file. Where a relative path is used it must be relative to the **exported file**.   

        :return: SVG <image> component
        :rtype: str
        """
        media_type = Path(self.path).suffix[1:]

        if self.embed:
            if media_type == 'svg':
                filepath = Path(self.path)
                with filepath.open() as f:
                    svg_data = f.read()
                return svg_group.render(
                    x = self.x,
                    y = self.y,
                    content = svg_data
                )
            else:
                encoded_img = base64.b64encode(open(self.path, "rb").read())
                path = 'data:image/{};base64,{}'.format(media_type, encoded_img.decode('utf-8'))
        else:
            path = self.path

        return svg_image.render(
            x = self.x,
            y = self.y,
            width = self.width,
            height = self.height,
            path = path
        )


class StyleSheet:
        def __init__(self, filepath, embed=False):
            """Include a stylesheet in the diagram

            :param filepath: Filename, including path, of the external stylesheet. *NOTE*: If *embedding*, a relative filepath is relative to the current working directory. If *linking*, a relative filepath is relative to the location of the final SVG diagram. 
            :type filepath: str
            :param embed: Elect to link or embed the stylesheet, defaults to False
            :type embed: bool, optional
            """
            self.filepath = filepath
            self.embed = embed

        def render(self):
            context = {}
            if self.embed:
                p = Path(self.filepath)
                context['css_data'] = p.read_text()
            else:
                context['filepath'] = self.filepath
            
            return svg_style.render(**context)


class Legend:

    ITEM_HEIGHT = 20
    ITEM_PAD = 4
    TEXT_PAD = 5
    SWATCH_PAD = 5
    INSET = 20

    def __init__(self, x, y, width, tags='', items=None):
        self.x = x
        self.y = y
        self.width = width
        self.items = items or []
        self.tags = tags
    
    @property
    def height(self):
        """Legend overall height (calculated dynamically).

        :return: Number of entries * preset item height
        :rtype: int
        """
        return len(self.items) * (self.ITEM_HEIGHT + self.ITEM_PAD) - self.ITEM_PAD + 2 * self.INSET

    @property
    def bounding_box(self):
        """Calculate a rectangular box that documents the bounds and location the rendered object 

        :return: namedTuple documenting x, y, width, and height.
        :rtype: diagram._BoundingBox
        """
        return _BoundingBox(self.x, self.y, self.width, self.height)

    def render(self):
        """Generates SVG code of the legend.

        :return: SVG legend component
        :rtype: str
        """
        # Parse user submitted item data
        items = []
        keys = ['name','tags','color']
        for i, values in enumerate(self.items):
            # Ensure values are a list (ie not a single item in a tuple)
            if isinstance(values, str):
                values = [values]

            item = dict(zip_longest(keys, values, fillvalue=''))
            
            # Create a pin and blank label as a 'swatch' for each item
            tags = ('swatch ' + item['tags']).strip()
            swatch = Pin(0, 0, -1, 0, [('', tags, 20, 20, 5)])
            item['swatch'] = swatch.render()

            item['x'] = swatch.width + self.INSET
            item['y'] = i * (self.ITEM_HEIGHT + self.ITEM_PAD) + (self.ITEM_HEIGHT / 2) + self.INSET
            items.append(item)
            

        return svg_legend.render(
            x = self.x,
            y = self.y,
            width = self.width,
            height = self.height,
            text_pad = self.TEXT_PAD,
            items = items,
            selectors = self.tags
        )



from pinout import diagram

# Create a new diagram
pinout_diagram = diagram.Diagram()

# Add a stylesheet
pinout_diagram.add_stylesheet('get_started_styles.css', embed=True)

# Add an image
pinout_diagram.add_image(0, 0, 220, 260, 'get_started_board.png', embed=True)

# Define some label categories/tags and add a legend
label_categories = [
    # (name, tag(s), color),
    ('GPIO', 'gpio'),
    ('Analog', 'analog'),
    ('PWM', 'pwm'),
    ('Power Management', 'pwr'),
]
pinout_diagram.add_legend(-160, 310, 225, 'legend legend-labels', label_categories)

# Set default values for labels
diagram.Label.default_width = 70
diagram.Label.default_height = 25
diagram.Label.default_gap = 6
diagram.Label.default_cnr = 3

# Add a pin and labels (the slow way)
pin01 = diagram.Pin(16, 100, -50, 100)

pin01.add_label('1', 'gpio')
pin01.add_label('A1', 'analog')
pin01.add_label('PWM', 'pwm')

pinout_diagram.components.append(pin01)

# A Pin and its labels can be added in a single line
# Note: this pin's coordinates target the left-most pin on the lower header. 
label_data = [('AREF', 'pwr')]
pinout_diagram.add_pin(65, 244, -50, 280, label_data)

# Add multiple pins and labels grouped as a list of 'header' dicts.
pin_headers = [
    {
        # LHS header - lower half
        'pin_coords': (16, 130),
        'label_coords': (-50 ,130),
        'pitch': 30,
        'labels': [
            [('Vcc', 'pwr')], 
            [('2', 'gpio'),('A2', 'analog')],
        ]
    },{
        # RHS header
        'pin_coords': (204, 100),
        'label_coords': (270 ,100),
        'pitch': 30,
        'labels': [
            [('8', 'gpio'),('A3', 'analog')], 
            [('7', 'gpio'),('A3', 'analog'), ('PWM','pwm')],
            [('GND', 'pwr')],
        ]
    },{
        # Lower header - remaining 3 pins
        'pin_coords': (95, 244),
        'label_coords': (270 ,280),
        'pitch': 30,
        'labels': [
            [('4', 'gpio'),('ADC', 'analog')], 
            [('5', 'gpio'),('ADC', 'analog'), ('PWM','pwm')],
            [('6', 'gpio'),('PWM', 'pwm', 70, 25, 82)],
        ]
    }
]

for header in pin_headers:
    pinout_diagram.add_pin_header(header)

#######################################
"""
# Add a Pin and add Labels (the fast way)
pinout_diagram.add_pin(16, 135, 'left', label_data)

# Add multiple pins and labels with some python-foo!
custom_specs = (70, 25, 50)
pin_label_data = [
        [('Vss', 'pwr', 40, 20, 200)], 
        [('GND', 'pwr', 40, 20, 200)], 
        [('#6', 'gpi',*custom_specs),('A3', 'analog'),('CLK', 'gpi')], 
        [('#5', 'gpio',*custom_specs),('A2', 'analog')], 
    ]

y_offset = 105
x_offset = 204
pitch = 30

for i, label_data in enumerate(pin_label_data):
    y = y_offset + pitch * i
    pinout_diagram.add_pin(x_offset, y, 'right', label_data)
"""
# Export the finished diagram
pinout_diagram.export('get_started_diagram.svg', overwrite=True)


import os
from pathlib import Path
from collections import namedtuple
from .templates import svg
from . import style_tools
from . import file_manager
from .components import Pin, Label, Legend, Image, StyleSheet, _BoundingBox, _Coords


class Diagram:
    """Components are collated and the final diagram is exported with this class. A typical diagram will include an image, pins with labels, and a stylesheet.
    """
    def __init__(self):
        self.components = []
        self.stylesheets = []

    def add_image(self, x, y, width, height, filepath, embed=False):
        """Create an image component and file it into the diagram in a single action. The SVG format officially supports JPG and PNG images. Other raster image formats may display unexpected behaviour. This component also accepts SVG files.

        :param x: Location of the image on the x axis
        :type x: int
        :param y: Location of the image on the y axis
        :type y: int
        :param width: Width of image in the diagram (may differ from actual image width)
        :type width: int
        :param height: Height of the image in the diagram (may differ from actual image height)
        :type height: int
        :param filepath: Filename, including path, to the image. **Note on relative paths**: When *linking*, the path must be relative to the exported SVG file location. When *embedding*, the path must be relative from the current working directory.
        :type filepath: string
        :param embed: Base64 encodes the image and embeds it in the SVG file, defaults to False
        :type embed: bool, optional
        """
        self.components.append(Image(x, y, width, height, filepath, embed))

    def add_stylesheet(self, filepath, embed=False):
        """Link an external stylesheet to the diagram. Multiple stylesheets can be added. They are referenced in the order added, this may be important where one style overrides another.

        :param filename: filename of stylesheet (include path to file)
        :type filename: str
        :param embed: Elect to embed the css file contents into the SVG file, defaults to False.
        :type embed: bool, optional
        """
        self.components.append(StyleSheet(filepath, embed))

    def add_pin(self, pin_x , pin_y, label_x, label_y, label_data=None):
        """Create a pin component, with associated labels, and file it into the diagram in a single action.

        :param pin_x: x-axis location of this pin
        :type pin_x: int
        :param pin_y: y-axis location of this pin
        :type pin_y: int
        :param label_x: x-axis location of the first label associated with this pin.
        :type label_x: int
        :param label_y: y-axis location of the first label associated with this pin.
        :type label_y: int
        :param label_data: A tuple with parameters required for Label(), defaults to None
        :type label_data: [type], optional
        """
        label_data = label_data or []
        pin = Pin(pin_x, pin_y, label_x, label_y, label_data)
        self.components.append(pin)

    def add_pin_header(self, pin_header):
        pin = _Coords(*pin_header['pin_coords'])
        lbl = _Coords(*pin_header['label_coords'])
        try:
            pitch = pin_header['pitch']
        except KeyError as e:
            pitch = 0
            if len(lbl) == 1:
                print('No \'pitch\' attribute present but \'label\' includes multiple lists.')
                print(pin_header)
                raise
        
        for i, lbl_data in enumerate(pin_header['labels']):
        
            # Calc coords for label locations (left & right)
            if lbl.y == pin.y:
                # Left or Right
                pin_ = _Coords(pin.x, pin.y + pitch * i)
                lbl_ = _Coords(lbl.x, lbl.y + pitch * i)
            elif lbl.y < pin.y:
                # Up
                if lbl.x < pin.x:
                    # Up Left 
                    pin_ = _Coords(pin.x + pitch * i, pin.y)
                    lbl_ = _Coords(lbl.x, lbl.y - pitch * i)
                else:
                    # Up Right
                    y_offset = lbl.y - (len(pin_header['labels']) - 1) * pitch
                    pin_ = _Coords(pin.x + pitch * i, pin.y)
                    lbl_ = _Coords(lbl.x, y_offset + pitch * i)
            else:
                # Down
                    if lbl.x < pin.x:
                        # Down Left
                        pin_ = _Coords(pin.x + pitch * i, pin.y)
                        lbl_ = _Coords(lbl.x, lbl.y + pitch * i)

                    else:
                        # Down Right
                        y_offset = lbl.y + (len(pin_header['labels']) - 1) * pitch
                        pin_ = _Coords(pin.x + pitch * i, pin.y)
                        lbl_ = _Coords(lbl.x, y_offset - pitch * i)
            

            self.add_pin(pin_.x, pin_.y, lbl_.x, lbl_.y, pin_header['labels'][i])
        

    def add_legend(self, x, y, width, tags, items):
        """Create a legend component and file it into the diagram in a single action.

        :param x: Location of the image on the x axis
        :type x: int
        :param y: Location of the image on the y axis
        :type y: int
        :param width: Width of the legend component. Set manually as font styling can unexpectedly affect content widths. 
        :type width: int
        :param tags: Applied to the legend as css class selector(s). Multiple tags can be included as a space separated list.
        :type tags: str, optional
        :param items: List of tuples documenting legend entries and associated tags. eg `[('GPIO', 'gpio'), ('GND', 'pwr-mgt')]`
        :type items: List
        """
        self.components.append(Legend(x, y, width, tags, items))

    def export(self, svgname, overwrite=False):
        """Output the diagram in SVG format. If no stylesheet(s) are included one will be generated and exported automatically. See style_tools.default_css() for more details.

        :param svgname: Name of svg file to be created, including export path.
        :type svgname: str
        :param overwrite: When set to False, this function aborts if the file already exists avoiding accidental overwriting. Defaults to False.
        :type overwrite: bool, optional
        """

        # Create export location and unique filename if required
        svgpath = Path(svgname)
        svgpath.parent.mkdir(parents=True, exist_ok=True)
        if not overwrite:
            svgpath = file_manager.unique_filepath(svgpath)
        svgpath.touch(exist_ok=True)

        styles = ''
        rendered_components = ''

        # Generate default styles if none supplied
        stylesheets = [s for s in self.components if isinstance(s, StyleSheet)]
        if not stylesheets:
            default_css = style_tools.default_css(self)
            cssname = svgpath.name[:-len(svgpath.suffix)]
            csspath = Path(svgpath.parent, '{}.css'.format(cssname))
            # File name may change if overwrite is False
            actual_csspath = file_manager.export_file(default_css, csspath, overwrite)
            
            self.components.insert(0, StyleSheet(actual_csspath.name))

        # Render all components
        for component in self.components:
            rendered_components += component.render()
        
        # Calculate viewbox to display all components
        viewbox_x = min([p.bounding_box.x for p in self.components if hasattr(p,'bounding_box')])
        viewbox_y = min([p.bounding_box.y for p in self.components if hasattr(p,'bounding_box')])
        viewbox_w = max([p.bounding_box.x + p.bounding_box.w for p in self.components if hasattr(p,'bounding_box')]) - viewbox_x
        viewbox_h = max([p.bounding_box.y + p.bounding_box.h for p in self.components if hasattr(p,'bounding_box')]) - viewbox_y
        
        # Render final SVG file
        svgpath.write_text(
            svg.render(
                x = 0,
                y = 0,
                width = viewbox_w,
                height = viewbox_h,
                viewbox = _BoundingBox(viewbox_x, viewbox_y, viewbox_w, viewbox_h),
                selectors = 'pinout-diagram',
                rendered_components = rendered_components,
                stylesheets = self.stylesheets,
                styles = styles or None
            )  
        )
        print(f'\'{svgpath}\' exported successfully.')


# File Manager: common file manipulation functions
from pathlib import Path

def unique_filepath(filepath):
    """Generate a unique filename to ensure existing files are not overwritten.

    :param filepath: File-name, including path, to the file location
    :type filepath: str
    :return: Unique filepath (filename and path to file location)
    :rtype: str
    """
    filepath = Path(filepath)
    suffix = filepath.suffix
    name = filepath.name[:-len(filepath.suffix)]
    path = filepath.parent
    count = 0

    while filepath.is_file():
        count += 1
        filepath = Path('{}/{}_{}{}'.format(path, name, count, suffix))
    return filepath


def export_file(content, path, overwrite=False):
    """Create a file and export content to it. If overwrite is False a new filename is generated.

    :param content: Text data to be written to file
    :type content: str
    :param path: Path, including filename, for the file to be created
    :type path: str
    :param overwrite: Dictate if an existing file should be overwritten. If False 'path' is amended to be a unique filename, defaults to False
    :type overwrite: bool, optional
    :return: Actual path the file was written to. Note: this may differ from supplied 'path' if 'overwrite' is False)
    :rtype: str
    """
    if isinstance(path, str):
        path = Path(path)
    if not overwrite:
        path = unique_filepath(path)
    path.touch(exist_ok=True)
    path.write_text(content)
    return path



import io
import random
import math
from pathlib import Path
from .components import Pin, Label
from .templates import stylesheet
from . import file_manager
    
def luminace(color_component):
    """Luminance of an individual Red, Green, or Blue, color component.

    :param color_component: Value between 0 and 255 (inclusive)
    :type color_component: int
    :return: Luminance value of the color component
    :rtype: float
    """
    i = float(color_component) / 255 

    if i < 0.03928:
        return i / 12.92
    else:
        return (( i + 0.055 ) / 1.055 ) ** 2.4
    
def relative_luminance(rgb):
    """Normalised luminance value of an RGB color.

    :param rgb: Tuple (or List) representing RGB value.
    :type rgb: tuple
    :return: Value between 0 and 1.
    :rtype: float
    """
    return 0.2126 * luminace(rgb[0]) + 0.7152 * luminace(rgb[1]) + 0.0722 * luminace(rgb[2]) 

def random_contrasting_rgb(ref_color):
    """Generate a psudo-random color that, compared to 'ref_color', has a contrast ratio greater than 3. This contrast value is the minimum value to pass WCAG AA contrast recommendation for UI components. 

    :param ref_color: Tuple (or List) representing RGB value.
    :type ref_color: tuple
    :return: Tuple representing an RGB color.
    :rtype: tuple
    """
    contrast = 0
    while contrast < 3:
        rgb = (random.randint(0, 255), random.randint(0, 255), random.randint(0, 255))
        lt = ref_color if sum(ref_color) > sum(rgb) else rgb
        dk = ref_color if sum(ref_color) < sum(rgb) else rgb
        contrast = ( relative_luminance(lt) + 0.05 ) / ( relative_luminance(dk) + 0.05 )
    return rgb

def default_css(diagram):
    """Generate a stylesheet using metrics from a diagram. Various styles are tailored by making a *best-guess* based on diagram component dimensions or a *lucky-guess* filtered by preset criteria. The output should be considered a boot-strapping step to styling a diagram ...unless you feel lucky!

    :param diagram: The Diagram object requiring styling
    :type diagram: Diagram
    :return: content of a css stylesheet with all required styles to display a diagram.
    :rtype: str
    """
    labels = [[l.tags.split(' ')[0] for l in c.labels] for c in diagram.components if isinstance(c, Pin)]
    labelset = set([tag for lbl in labels for tag in lbl])
    
    label_font_size = math.floor(Label.default_height * (3/5))
    label_text_color = (255,255,255)
    
    return stylesheet.render(
        legend = [(l.capitalize(), l, 'rgb' + str(random_contrasting_rgb(label_text_color))) for l in labelset],
        legend_font_size = max(13, label_font_size),
        label_font_size = label_font_size,
        label_text_color = 'rgb' + str(label_text_color),
    )


import pkg_resources

def duplicate(resource_name='get started'):
    """pinout includes some sample projects. These can be duplicated to the working directory and used in conjunction with the official tutorial.

    Current projects:

        + 'get started'
        + 'full sample'

    :param resource_name: Name of sample project. Defaults to 'get started'.
    :type resource_name: string
    """

    resources = {
        'get started': [
            ('get_started','get_started_pinout.py'),
            ('get_started','get_started_board.png'),
            ('get_started','get_started_styles.css')
        ],
        'full sample':[
            ('sample_full','sample_full_pinout.py'),
            ('sample_full','sample_full_pinout_hardware.png'),
            ('sample_full','sample_full_pinout_styles.css')
        ]
    }

    resource_package = __name__
    for path in resources[resource_name]:
        resource_path = '/'.join(('resources', *path))
        data = pkg_resources.resource_string(resource_package, resource_path)
        filename = path[-1]
        with open(filename, 'wb') as f:
            f.write(data)
        print(f'{filename} duplicated.')

        